function [gBestX,gBestCost,bestMdl] = pso(fx,Lb,Ub,MaxPop,MaxIters,itersToReset)
% PSO implements particle swarm optimization.
%
% INPUT
% ======
% fx - handle of optimized function.
% Lb - array of low bound values.
% Ub - array of upper bound values.
% MaxPop - maximum population of swarm.
% MaxIters - maxumum number of iterations
% itersToReset - number of stalled iterations that lead to data reset.
%
% OUTPUT
% ======
% gBestX - array of best solutions.
% gBestCost - best optimized function value.
%
% Example
% =======
%
% >> [gBestX,gBestCost] = pso(@fx2,[0 0 0 0],[5 5 5 5],50,500)
%
  global mdl

  nVar = length(Lb);
  template.position = zeros(1,nVar);
  template.velocity = zeros(1,nVar);
  template.cost =  0;
  template.best_position = zeros(1,nVar);
  template.best_cost = 1e+99;
  
  c1 = 1.5;
  c2 = 2;
  w = 1;
  wdamp = 0.995;
  gBestCost = 1;
  gBestX = zeros(1,nVar);

  % Initizialize population
  for i=1:MaxPop
    pop(i) = template;
    for j=1:nVar
      pop(i).position(j) = fix(Lb(j) + (Ub(j) - Lb(j))*rand);
      pop(i).velocity(j) = 0;
    end
    
    pop(i).cost = fx(pop(i).position);
    
    if pop(i).cost < pop(i).best_cost
      pop(i).best_cost = pop(i).cost; 
      pop(i).best_position = pop(i).position;

    end 
    
    if pop(i).best_cost < gBestCost
      bestMdl = mdl;
      gBestCost = pop(i).best_cost;
      gBestX = pop(i).best_position;
    end 
  end

  fprintf('Best Fx = %e, ', gBestCost);  
  fprintf('Best X =[');
  fprintf('%d, ', gBestX(1:nVar-1));
  fprintf(' %d]\n', gBestX(nVar));
  lastIter = 0;
  % pso loop
  for iter = 1:MaxIters
    if gBestCost == 0,return; end
    if (iter - lastIter) >= itersToReset
      fprintf("----------------------- Reset population ----------------------- \n")
      fprintf("AT iteration %d\n", iter);
      lastIter = iter;
      for i=1:MaxPop
        pop(i) = template;
        for j=1:nVar
          pop(i).position(j) = fix(Lb(j) + (Ub(j) - Lb(j))*rand);
          pop(i).velocity(j) = 0;
        end
        
        pop(i).cost = fx(pop(i).position);
        
        if pop(i).cost < pop(i).best_cost
          pop(i).best_cost = pop(i).cost; 
          pop(i).best_position = pop(i).position;
        end 
        
        if pop(i).best_cost < gBestCost
          lastIter = iter;
          bestMdl = mdl;
          gBestCost = pop(i).best_cost;
          gBestX = pop(i).best_position;         
          fprintf('Iter = %i, ', iter);
          fprintf('Best Fx = %e, ', gBestCost); 
          fprintf('Best X = [');
          fprintf('%d, ', gBestX(1:nVar-1));
          fprintf(' %d]\n', gBestX(nVar));
        end 
      end      
    end

    for i = 1:MaxPop
      for j = 1:nVar
        pop(i).velocity(j) = w * pop(i).velocity(j) + ...
          c1 * rand * (pop(i).best_position(j) - pop(i).position(j)) + ...
          c2 * rand * (gBestX(j) - pop(i).position(j));
        pop(i).position(j) = fix(pop(i).position(j) + pop(i).velocity(j));
        if pop(i).position(j) < Lb(j) || pop(i).position(j) > Ub(j)
          pop(i).position(j) = fix(Lb(j) + (Ub(j) - Lb(j)) * rand);
        end 
      end
      
      pop(i).cost = fx(pop(i).position);
      
      if pop(i).cost < pop(i).best_cost 
        pop(i).best_cost = pop(i).cost;
      end

      if pop(i).best_cost < gBestCost
        lastIter = iter;
        bestMdl = mdl;
        gBestCost = pop(i).best_cost;
        gBestX = pop(i).best_position;
        fprintf('Iter = %i, ', iter);
        fprintf('Best Fx = %e, ', gBestCost);
        fprintf('Best X = [');
        fprintf('%d, ', gBestX(1:nVar-1));
        fprintf(' %d]\n', gBestX(nVar));
      end
    end
    w = w * wdamp;
  end
end

